; Solid state 6U5 magic eye tube replacement
; 
; R. Weaver, 2014
;
; I/O assignment:
; Port A.0 - Analog input
; LED outputs, LED is a 4x4 matrix of 16 LEDs
; 4 anode lines A0..A3, and 4 cathode lines K0..K3
; Port A.1 - digital out LED A0
; Port A.2 - digital out LED A1
; POrt C.0 - digital out LED A2
; POrt C.1 - digital out LED A3
; POrt C.2 - digital out LED K3
; POrt C.3 - digital out LED K2
; POrt C.4 - digital out LED K1
; POrt C.5 - digital out LED K0

;Notes
;start to fade in after 15 seconds. Fade in complete after 20 seconds.

    #include p16f684.inc
    __config _FOSC_EC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_OFF & _BOREN_OFF & _CP_OFF & _CPD_OFF
;
; Miscellaneous macros. May not all be used in this program
;
skpzer macro
    btfss status,z
    endm
skpeq macro
    btfss status,z
    endm
skpnzr macro
    btfsc status,z
    endm
skpneq macro
    btfsc status,z
    endm
skpneg macro
    btfsc status,c
    endm
skppos macro
    btfss status,c
    endm
bank0 macro
    bcf status,RP0
    endm
bank1 macro
    bsf status,RP0
    endm
dde macro arg ;Store a 16 bit value in adjacent memory locs
    de low arg
    de high arg
    endm
;
;    Equates
eebase equ 0x2100 ;base address of EEPROM
; Misc constants
warmupdelay equ d'240' ; tube warmup time
fadeintime equ d'255' ; fade in time after warmup time
fadeintime2 equ d'8' ; fade in time after warmup time

;    Variables
;
    cblock 0x20    ; start of user ram
    temp1,temp2 ; misc use
    count ;misc counter register
    LEDoffset ;offset pointer for LED lookup table
    FadeInFlag ; bit 0 set to indicate fade-in in progress
    FadeInCtrL
    dispVal ;scaled analog value to be displayed
    dutyCycleCtr,dutycycle,countAdj ;for dithering display
    dlyctrH,dlyctrL,ShortDelay ;delay routine registers
;   wsave,statsave ;interrupt save registers
    endc
;
;
; ********************************************
;
;
    code    0x00
Reset
    goto Main    ; Entry point for program from Reset
    nop
    nop
    nop
;    goto    IntHndl    ; Vector for interrupt routine
    goto Main ;if an unintended interrupt happens, simply reboot.
; LED lookup table ordered for bargraph application
LookupBGraph
;    andlw 0x1F
    addwf pcl,f
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0x00
    retlw 0xE8
    retlw 0x71
    retlw 0xE4
    retlw 0x72
    retlw 0xE2
    retlw 0x74
    retlw 0xE1
    retlw 0x78
    retlw 0xD8
    retlw 0xB1
    retlw 0xD4
    retlw 0xB2
    retlw 0xD2
    retlw 0xB4
    retlw 0xD1
    retlw 0xB8
    retlw 0x00
    retlw 0x00
    retlw 0x00
LookupDly ;A table of exponentially decaying values
    andlw 0x0F
    addwf pcl,f
    retlw d'1'
    retlw d'2'
    retlw d'3'
    retlw d'4'
    retlw d'6'
    retlw d'8'
    retlw d'11'
    retlw d'16'
    retlw d'23'
    retlw d'32'
    retlw d'45'
    retlw d'64'
    retlw d'90'
    retlw d'128'
    retlw d'180'
    retlw d'255'
;
Main
;Initialize the ports etc.
    bank1
    clrf intcon ;initially disable interrupts
    clrf iocA
    movlw b'00111001'
    movwf trisA
    clrf trisC
    movlw b'01110000' ;use internal ADC clock
    movwf adcon1
    clrf ansel
    bsf ansel,0 ;portA,0 is analog in

    bank0
    movlw b'00000001' ;select left justified analog result
    movwf adcon0
    clrf dutyCycleCtr
    clrf countAdj
    clrf portA
    clrf portC
    bsf adcon0,go ;start analog conversion
    call WarmUp
    bsf FadeInFlag,0
    movlw fadeintime2
    movwf FadeInCtrL
    clrf dispVal
MainLoop
;
;
bargraphloop1
    bcf status,c
    swapf dispVal,w ;get value to be displayed
    andlw 0x0E
    movwf LEDoffset
    movf dispVal,w
    andlw 0x1F
    movwf dutycycle
    movlw 0x12
    movwf count
bargraphloop2 ; call up 16 contiguous entries from the LED lookup table
    movf LEDoffset,w ; get the main offset
    addwf countAdj,w ; add in the dither bit
    call LookupBGraph ;
    movwf temp1
    rrf temp1,f ;split the data into portA and portC bytes
    rrf temp2,f
    rrf temp1,f
    rrf temp2,f
    rrf temp2,f
    swapf temp2,f
    ; the following output transfer sequence prevents
    ; random LEDs from being briefly switched on
    clrf portA
    movf temp1,w
    movwf portC
    movf temp2,w
    movwf portA
    incf LEDoffset,f
    decfsz count,f
    goto bargraphloop2
; Duty cycle adjustment
; This dithers adjacent LEDs with time weighting according to
; the low six bits of dispVal, resulting in a smooth transition
; from one LED to the next.
DcycAdj
    incf dutyCycleCtr,f
    movlw 0x1F
    andwf dutyCycleCtr,f
dutyCyclAdjust
    bcf countAdj,1
    movf dutycycle,w
    subwf dutyCycleCtr,w
    skppos
    bsf countAdj,1
    clrf portA
    clrf portC
    comf portC,f
ReadADC
    ;by now the adc conversion should be complete
    btfsc adcon0,go
    goto bargraphloop1
    call GetAnalog
    btfsc FadeInFlag,0
    goto FadeIn
    movwf dispVal
    goto EndADCread
FadeIn
    decfsz FadeInCtrL,f
    goto EndADCread
    movlw fadeintime2
    movwf FadeInCtrL
    incf dispVal,f ;ramp up the display value
    call GetAnalog
    subwf dispVal,w ;see if it has reached the actual analog value
    skppos
    goto bargraphloop1
    bcf FadeInFlag,0 ; fade in is complete
    decf dispVal,f
EndADCread
    bsf adcon0,go ;restart adc
    goto bargraphloop1
; Start of subroutines
GetAnalog
    ;analog is in range of 1.6..5 volts
    ;with 5V= 0 counts and 1.6 = 255 counts approx.
    comf adresh,w ;negate the input
    movwf temp1
    bcf status,c
    rrf temp1,w ;get half the value
    addwf temp1,w ;add to the whole value to get 1.5x
    btfsc status,c ;if overflow set to max value
    movlw 0xFF
    return
Warmup
    clrf portA
    clrf portC
    movlw warmupdelay ; about 10 second delay to
    call LongDelay ;simulate tube warmup time
    ; now fade in the zero LED
    ; with portA bits all off, the zero LED can
    ; be turned on and off simply by controlling portC.
    ; zero LED off: portC = 00
    ; zero LED on: portC  = 3A
    clrf dutyCycleCtr
    movlw fadeintime
    movwf FadeInCtrL
    movlw 0x38 ; Set cathode line 0 low & other cathode lines high
    movwf portC
    bcf portC,1 ;turn off zero LED
    movlw 0x10
    movwf dutycycle
zeroLEDloop
    decf dutycycle,w ;load dutycycle minus one
    call LookupDly
    movwf dutyCycleCtr
    bsf portC,1 ; blip the LED
    nop
    bcf portC,1
    clrf temp1
    bsf status,c
zeroLEDlp2
    movlw 0x06
    movwf temp1
    decfsz temp1,f
    goto $-1
    decfsz dutyCycleCtr,f
    goto zeroLEDlp2
    decfsz FadeInCtrL,f
    goto zeroLEDloop
    movlw fadeintime
    movwf FadeInCtrL
    bcf status,c
    decfsz dutycycle,f
    goto zeroLEDloop
FadeInLeds
    bsf adcon0,go ;restart adc
    btfsc adcon0,go
    goto $-1 ;wait for ADC
    return

LongDelay
    movwf count
    movlw ShortDelay
LdelayLp
    call delay
    decfsz count,f
    goto LdelayLp
    return
delay
    ; 00002 cyc for entry call
    movwf    dlyctrH    ; 00001 cyc, w contains delay value
subloop
    decfsz    dlyctrL,f    ; 00768 cyc for loop * w
    goto $-1    ; -0001 cyc for loop exit * w
    decfsz dlyctrH,f    ; 00001 * w
    goto subloop    ; 00002 * w
    ; -0001 for loop exit
    return ; 00002
;   EEramData
    org 0x2100
Eoffset dde d'987' ;Initial offset value
    end
